library webview;

import 'dart:async';
import 'dart:convert';
import 'dart:html';
import 'dart:js' as js;

class Webview {

  static const Map _EVENT_DETAIL = const { 
    'exit' : const ['processId', 'reason'],
    'loadabort' : const ['url', 'isTopLevel', 'reason'],
    'loadcommit' : const ['url', 'isTopLevel'],
    'loadredirect' : const ['oldUrl', 'newUrl', 'isTopLevel'],
    'loadstart' : const ['url', 'isTopLevel'],
    'loadstop' : const [],
    'sizechanged': const ['oldHeight', 'oldWidth', 'newHeight', 'newWidth']
  };
    
  /// The url of the embedded content.
  String get src => _webview['src'];
         set src(String value) => _webview['src'] = value;
         
  /// Whether or not this webview can navigate [back].
  bool get canGoBack => _webview.callMethod('canGoBack');
  
  /// Whether or not this webview can navigate [forward].
  bool get canGoForward => _webview.callMethod('canGoForward');
  
  HtmlWindow _contentWindow;
  /// The window object of this webview.
  HtmlWindow get contentWindow {
    if(_contentWindow == null) {
      _contentWindow = new HtmlWindow(_webview.callMethod('contentWindow'));
    }
    return _contentWindow;
  }

  final _onExit = new StreamController<CustomEvent>.broadcast();
  /// Dispatched when the process rendering the webview content has exited.
  ///
  /// Details:
  ///   `int processId`:        The [processId] of the webview that exited.
  ///   `String reason`:        One of `normal`, `abnormal`, `crash`, `kill`.
  Stream<CustomEvent> get onExit => _onExit.stream;
  
  final _onLoadAbort = new StreamController<CustomEvent>.broadcast();
  /// Dispatched when a top-level load has aborted without committing.
  ///
  /// Details:
  ///   `String url`:           The requested url.
  ///   `boolean isTopLevel`:   Whether the load was top-level or in a subframe.
  ///   `String reason`:        One of `networkError`, `download`, `canceled`,
  ///                           `sslError`, or `safeBrowsingError`.
  Stream<CustomEvent> get onLoadAbort => _onLoadAbort.stream;
  
  final _onLoadCommit = new StreamController<CustomEvent>.broadcast();
  /// Dispatched when a load has committed.
  ///
  /// Details:
  ///   `String url`:           The url that committed.
  ///   `boolean isTopLevel`:   Whether the load is top-level or in a subframe.
  Stream<CustomEvent> get onLoadCommit => _onLoadCommit.stream;
  
  final _onLoadRedirect = new StreamController<CustomEvent>.broadcast();
  /// Dispatched when a top-level load request has redirected to another url.
  ///
  /// Details:
  ///   `String oldUrl`:        The requested url, before the redirect.
  ///   `String newUrl`:        The new requested url, after the redirect.
  ///   `boolean isTopLevel`:   Whether the redirect is top-level or in a 
  ///                           subframe.
  Stream<CustomEvent> get onLoadRedirect => _onLoadRedirect.stream;
  
  final _onLoadStart = new StreamController<CustomEvent>.broadcast();
  /// Dispatched when a load has begun.
  ///
  /// Details:
  ///   `String url`:           The requested url.
  ///   `boolean isTopLevel`:   Whether the load is top-level or in a subframe.
  Stream<CustomEvent> get onLoadStart => _onLoadStart.stream;
  
  final _onLoadStop = new StreamController<CustomEvent>.broadcast();
  /// Dispatched when all loads in the element (including all subframes) have 
  /// completed. This event is dispatched in addition to any [loadCommitEvent] 
  /// or [loadAbortEvent], which occur per-frame.
  Stream<CustomEvent> get onLoadStop => _onLoadStop.stream;
  
  /// The process id of this webview.
  int get processId => _webview.callMethod('getProcessId');

  final Element element;  
  js.JsObject _webview;

  Webview() : this.wrap(new Element.tag('webview'));
  
  Webview.wrap(this.element) {
    _webview = new js.JsObject.fromBrowserObject(element);
    _webview.callMethod('addEventListener', ['exit', _dispatch]);
    _webview.callMethod('addEventListener', ['loadabort', _dispatch]);
    _webview.callMethod('addEventListener', ['loadcommit', _dispatch]);
    _webview.callMethod('addEventListener', ['loadredirect', _dispatch]);
    _webview.callMethod('addEventListener', ['loadstart', _dispatch]);
    _webview.callMethod('addEventListener', ['loadstop', _dispatch]);
  }
  
  void dispose() {
    // TODO(rms): close listeners
  }
  
  /// Navigate back one step in this webview's history.
  void back() => _webview.callMethod('back');  
  
  /// Inject and execute a script in the webview's context.
  /// 
  /// NOTE: this is an experimental and undocumented feature.  It only works at
  /// present with javascript, and you should specify one of [code] or [file],
  /// but _not_ both.
  void executeScript({String code, String file}) {
    final m = {};
    if (code != null) m['code'] = code;
    if (file != null) m['file'] = file;
    _webview.callMethod('executeScript', [new js.JsObject.jsify(m)]);
  }
  
  /// Navigate forward one step in this webview's history.
  void forward() => _webview.callMethod('forward');
  
  /// Navigate to a [relativeIndex] in this webview's history.
  ///
  /// If the [relativeIndex] is positive it means to go forward that number of
  /// history entries; if negative it means to go back that number of entries.
  void go(int relativeIndex) => _webview.callMethod('go', [relativeIndex]);
  
  /// Reload the content of this webview.
  void reload() => _webview.callMethod('reload');
  
  /// Stops this webview if it is currently loading content.
  void stop() => _webview.callMethod('stop');  
  
  /// Terminates the content's renderer process.
  ///
  /// This may affect multiple webviews in the current app if they share 
  /// [processId], but it will not affect webviews in other apps.
  void terminate() => _webview.callMethod('terminate');
    
  bool _dispatch(e) {    
    e = new js.JsObject.fromBrowserObject(e);    
    var detail = new Map();
    for(var d in _EVENT_DETAIL[e['type']]) {      
      switch(d) {
        case 'isTopLevel' : detail['isTopLevel'] = e['isTopLevel']; break;  
        case 'oldHeight' : detail['oldHeight'] = e['oldHeight']; break;
        case 'oldUrl' : detail['oldUrl'] = e['oldUrl']; break;
        case 'oldWidth' : detail['oldWidth'] = e['oldWidth']; break;
        case 'newHeight' : detail['newHeight'] = e['newHeight']; break;
        case 'newUrl' : detail['newUrl'] = e['newUrl']; break;
        case 'newWidth' : detail['newWidth'] = e['newWidth']; break;
        case 'processId' : detail['processId'] = e['processId']; break;
        case 'reason' : detail['reason'] = e['reason']; break;
        case 'url': detail['url'] = e['url']; break;
      }
    }    
    final ev = new CustomEvent(
        e['type'], 
        canBubble: e['bubbles'], 
        cancelable: e['cancelable'], 
        detail: JSON.encode(detail));    
    switch(ev.type) {
      case 'exit': _onExit.add(ev); break;
      case 'loadabort': _onLoadAbort.add(ev); break;
      case 'loadcommit': _onLoadCommit.add(ev); break;
      case 'loadredirect': _onLoadRedirect.add(ev); break;
      case 'loadstart': _onLoadStart.add(ev); break;
      case 'loadstop': _onLoadStop.add(ev); break;
      default: break;
    }
  }
}

class HtmlWindow {
  js.JsObject _proxy;  
  HtmlWindow(this._proxy);
  void postMessage(message, String targetOrigin) => 
      _proxy.callMethod('postMessage', [message, targetOrigin]);
}
